"""Allowed windows units and lengths in seconds."""

ALLOWED_WINDOWS = {
    "second": 1,
    "seconds": 1,
    "minute": 60,
    "minutes": 60,
    "hour": 3600,
    "hours": 3600,
    "day": 86400,
    "days": 86400,
    "week": 604800,
    "weeks": 604800,
    "year": 29030400,
    "years": 29030400,
}
